from typing import List
from pydantic import BaseModel
from typing import TYPE_CHECKING

from ponika.models import ApiResponse

if TYPE_CHECKING:
    from ponika import PonikaClient, ApiResponse


class MessagesEndpoint:
    def __init__(self, client: "PonikaClient") -> None:
        self.client: "PonikaClient" = client
        self.actions = self.MessagesActionsEndpoint(client)

    class MessagesResponseDataItem(BaseModel):
        """Data model for messages response."""

        message: str
        sender: str
        id: str
        modem_id: str
        status: str
        date: str

    def get_status(self) -> "ApiResponse[List[MessagesResponseDataItem]]":
        """Fetch messages from the device."""
        return ApiResponse[List[self.MessagesResponseDataItem]].model_validate(
            self.client._get(
                "/messages/status",
            )
        )

    class MessagesActionsEndpoint:
        def __init__(self, client: "PonikaClient") -> None:
            self.client: "PonikaClient" = client

        class SendResponseData(BaseModel):
            """Data model for messages action response."""

            sms_used: int

        def post_send(
            self, number: str, message: str, modem: str
        ) -> "ApiResponse[SendResponseData]":
            """Send a message to a recipient."""
            return self.client._post(
                "/messages/actions/send",
                self.SendResponseData,
                {"data": {"number": number, "message": message, "modem": modem}},
            )



from typing import List, Literal, Optional
from pydantic import BaseModel
from typing import TYPE_CHECKING

from ponika.models import ApiResponse

if TYPE_CHECKING:
    from ponika import PonikaClient


class TailscaleEndpoint:
    def __init__(self, client: "PonikaClient") -> None:
        self.client: "PonikaClient" = client

    class TailscaleConfigResponseDataItem(BaseModel):
        """Data model for Tailscale configuration response."""

        id: Optional[str] = None
        enabled: Optional[str] = None
        auth_key: Optional[str] = None
        advert_routes: Optional[List[str]] = None
        accept_routes: Optional[str] = None
        exit_node: Optional[str] = None
        auth_type: Literal["url", "key"]
        default_route: Optional[Literal["0", "1"]] = None
        exit_node_ip: Optional[str] = None
        login_server: Optional[str] = None

    def get_config(self) -> "ApiResponse[List[TailscaleConfigResponseDataItem]]":
        """Fetch Tailscale configuration from the device."""
        return ApiResponse[List[self.TailscaleConfigResponseDataItem]].model_validate(
            self.client._get("/tailscale/config")
        )

    class TailscaleStatusResponseDataItem(BaseModel):
        """Data model for Tailscale status response."""

        status: str
        url: str
        ip: List[str]
        message: List[str]

    def get_status(self) -> "ApiResponse[List[TailscaleStatusResponseDataItem]]":
        """Fetch Tailscale status from the device."""
        return ApiResponse[List[self.TailscaleStatusResponseDataItem]].model_validate(
            self.client._get("/tailscale/status")
        )



from typing import Optional
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING

from ponika.models import ApiResponse

if TYPE_CHECKING:
    from ponika import PonikaClient


class DhcpEndpoint:
    def __init__(self, client: "PonikaClient") -> None:
        self.client: "PonikaClient" = client
        self.leases = self.DhcpLeasesEndpoint(client)

    class DhcpLeasesEndpoint:
        def __init__(self, client: "PonikaClient") -> None:
            self.client: "PonikaClient" = client
            self.ipv4 = self.DhcpLeasesIpv4Endpoint(client)

        class DhcpLeasesIpv4Endpoint:
            def __init__(self, client: "PonikaClient") -> None:
                self.client: "PonikaClient" = client

            class DhcpLeasesIpv4ResponseDataItem(BaseModel):
                """Data model for IPv4 DHCP leases response."""

                expires: Optional[int] = Field(
                    description="Amount of time left for DHCP lease to expire."
                )
                macaddr: Optional[str] = Field(description="MAC address of the device.")
                ipaddr: Optional[str] = Field(
                    description="IP address assigned to the device."
                )
                hostname: Optional[str] = Field(
                    default=None, description="Hostname of the device, if available."
                )
                interface: Optional[str] = Field(
                    default=None,
                    description="Network interface the device is connected to.",
                )

            def get_status(
                self,
            ) -> "ApiResponse[list[DhcpLeasesIpv4ResponseDataItem]]":
                """Fetch IPv4 DHCP leases from the device."""
                return ApiResponse[
                    list[self.DhcpLeasesIpv4ResponseDataItem]
                ].model_validate(self.client._get("/dhcp/leases/ipv4/status"))



from pydantic import BaseModel
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ponika import PonikaClient, ApiResponse


class UnauthorizedEndpoint:
    def __init__(self, client: "PonikaClient") -> None:
        self.client: "PonikaClient" = client

    class UnauthorizedStatusResponseData(BaseModel):
        """Data model for GET /unauthorized/status response."""

        device_name: str
        device_model: str
        device_identifier: str
        api_version: str
        lang: str

    def get_status(self) -> "ApiResponse[UnauthorizedStatusResponseData]":
        """Fetch unauthorized status from the device."""
        self.client.logger.info("Accessing unauthorized endpoint...")
        return ApiResponse[self.UnauthorizedStatusResponseData].model_validate(
            self.client._get(
                "/unauthorized/status",
            )
        )



from pydantic import BaseModel
from typing import TYPE_CHECKING

from ponika.models import ApiResponse

if TYPE_CHECKING:
    from ponika import PonikaClient


class SessionEndpoint:
    def __init__(self, client: "PonikaClient") -> None:
        self.client: "PonikaClient" = client

    class SessionResponseData(BaseModel):
        """Data model for session response."""

        active: bool

    def get_status(self) -> "ApiResponse[SessionResponseData]":
        """Fetch session information from the device."""

        return ApiResponse[self.SessionResponseData].model_validate(
            self.client._get("/session/status")
        )



from typing import Type, Optional, Dict, Any
from requests import Session
from logging import Logger, getLogger
from pydantic import BaseModel, validate_call
from time import time

from ponika.endpoints.dhcp import DhcpEndpoint
from ponika.endpoints.gps import GpsEndpoint
from ponika.endpoints.internet_connection import InternetConnectionEndpoint
from ponika.endpoints.messages import MessagesEndpoint
from ponika.endpoints.session import SessionEndpoint
from ponika.endpoints.tailscale import TailscaleEndpoint
from ponika.endpoints.unauthorized import UnauthorizedEndpoint
from ponika.endpoints.wireless import WirelessEndpoint
from ponika.models import T, ApiResponse, Token


class PonikaClient:
    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        port: int | None = None,
        tls: bool = True,
        verify_tls: bool = True,
    ) -> None:
        self.username = username
        self.password = password
        self.host = host
        self.port = port or (443 if tls else 80)
        self.tls = tls
        self.verify_tls = verify_tls
        self.base_url = f"{'https' if tls else 'http'}://{self.host}:{self.port}/api"

        self.request: Session = Session()
        self.logger: Logger = getLogger(__name__)

        self.auth: None | Token = None

        self.unauthorized = UnauthorizedEndpoint(self)
        self.session = SessionEndpoint(self)
        self.messages = MessagesEndpoint(self)
        self.gps = GpsEndpoint(self)
        self.dhcp = DhcpEndpoint(self)
        self.tailscale = TailscaleEndpoint(self)
        self.wireless = WirelessEndpoint(self)
        self.internet_connection = InternetConnectionEndpoint(self)

    def get_auth_token(self) -> Optional[str]:
        """Get the current authentication token."""
        if self.auth and self.auth.expires_at > int(time()):
            return self.auth.token

        auth_response = self.login(self.username, self.password)

        self.auth = (
            Token(
                token=auth_response.data.token,
                expires_at=int(time()) + auth_response.data.expires,
            )
            if auth_response.success and auth_response.data
            else None
        )

        return self.auth.token if self.auth else None

    def _get(
        self,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        auth_required: bool = True,
    ) -> object:
        self.logger.info("Making GET request to: %s", endpoint)

        auth_token = self.get_auth_token() if auth_required else None

        response = self.request.get(
            f"{self.base_url}{endpoint}",
            verify=self.verify_tls,
            params=params,
            headers=({"Authorization": f"Bearer {auth_token}"} if auth_token else None),
        )

        return response.json()

    def _post(
        self,
        endpoint: str,
        data_model: Type[T],
        params: Optional[Dict[str, Any]] = None,
        auth_required: bool = True,
    ) -> ApiResponse[T]:
        self.logger.info("Making POST request to: %s", endpoint)

        auth_token = self.get_auth_token() if auth_required else None

        response = self.request.post(
            f"{self.base_url}{endpoint}",
            verify=self.verify_tls,
            json=params,
            headers=({"Authorization": f"Bearer {auth_token}"} if auth_token else None),
        )

        return ApiResponse[data_model].model_validate(response.json())

    class LoginResponseData(BaseModel):
        """Data model for login response."""

        username: str
        token: str
        expires: int

    @validate_call
    def login(self, username: str, password: str) -> ApiResponse[LoginResponseData]:
        """Login to the Ponika API and retrieve a token."""
        self.logger.info("Logging in with username: %s", username)
        response = self._post(
            "/login",
            self.LoginResponseData,
            {"username": username, "password": password},
            auth_required=False,
        )

        return response

    class LogoutResponseData(BaseModel):
        """Data model for logout response."""

        response: str

    def logout(self) -> ApiResponse[LogoutResponseData]:
        """Logout from the Ponika API."""
        self.logger.info("Logging out...")
        return self._post("/logout", self.LogoutResponseData)



from typing import Generic, List, Optional, TypeVar
from pydantic import BaseModel


T = TypeVar("T")


class TeltonikaApiError(BaseModel):
    """Custom exception for Teltonika API errors."""

    code: int
    error: str
    source: str
    section: Optional[str] = None

    def __str__(self) -> str:
        return f"Error {self.code}: {self.error} ({self.source}, {self.section})"


class ApiResponse(BaseModel, Generic[T]):
    success: bool
    data: None | T = None
    errors: Optional[List[TeltonikaApiError]] = None


class Token(BaseModel):
    """Data model for token storage."""

    token: str
    expires_at: int



import os
from ponika import PonikaClient
from os import environ
import pytest

TELTONIKA_HOST = environ.get("TELTONIKA_HOST")
TELTONIKA_USERNAME = environ.get("TELTONIKA_USERNAME")
TELTONIKA_PASSWORD = environ.get("TELTONIKA_PASSWORD")


def create_client():
    if not TELTONIKA_HOST or not TELTONIKA_USERNAME or not TELTONIKA_PASSWORD:
        raise ValueError(
            "Environment variables TELTONIKA_HOST, TELTONIKA_USERNAME, and TELTONIKA_PASSWORD must be set."
        )

    """Create a PonikaClient instance for testing."""
    return PonikaClient(
        TELTONIKA_HOST, TELTONIKA_USERNAME, TELTONIKA_PASSWORD, verify_tls=False
    )


def test_client_logout():
    """Test the logout functionality of the PonikaClient."""

    logout_response = create_client().logout()
    assert logout_response.success


def device_doesnt_support_skipper(response):
    """Check if the device supports a specific feature based on the response."""
    if not response.success and response.errors[0].code == 122:
        # This indicates the device doesn't have the endpoint so we skip the test
        pytest.skip("Endpoint not available on this device.")
        # raise ValueError("Response is not successful or data is missing.")
    return response


def test_client_session_status():
    """Test the session status functionality of the PonikaClient."""

    session_status_response = device_doesnt_support_skipper(
        create_client().session.get_status()
    )

    assert session_status_response.success


def test_client_gps_get_global():
    """Test the GPS global functionality of the PonikaClient."""

    gps_global_response = device_doesnt_support_skipper(
        create_client().gps.get_global()
    )
    assert gps_global_response.success


def test_client_gps_get_status():
    """Test the GPS status functionality of the PonikaClient."""

    gps_status_response = device_doesnt_support_skipper(
        create_client().gps.get_status()
    )
    assert gps_status_response.success


def test_client_gps_position_get_status():
    """Test the GPS status functionality of the PonikaClient."""

    gps_status_response = device_doesnt_support_skipper(
        create_client().gps.position.get_status()
    )
    assert gps_status_response.success


def test_client_messages_get_status():
    """Test the messages status functionality of the PonikaClient."""

    messages_status_response = device_doesnt_support_skipper(
        create_client().messages.get_status()
    )
    assert messages_status_response.success


@pytest.mark.skipif(
    os.getenv("MOBILE_NUMBER") is None,
    reason="Mobile number ($MOBILE_NUMBER) required for this test",
)
def test_client_messages_actions_post_send():
    """Test the messages actions send functionality of the PonikaClient."""

    messages_actions_send_response = device_doesnt_support_skipper(
        create_client().messages.actions.post_send(
            number=str(os.getenv("MOBILE_NUMBER")), message="Hello, World!", modem="2-1"
        )
    )
    assert messages_actions_send_response.success


def test_client_dhcp_leases_ipv4_get_status():
    """Test the leases IPv4 status functionality of the PonikaClient."""

    leases_dhcp_leases_ipv4_sg = device_doesnt_support_skipper(
        create_client().dhcp.leases.ipv4.get_status()
    )
    assert leases_dhcp_leases_ipv4_sg.success


def test_client_tailscale_get_config():
    """Test the Tailscale configuration functionality of the PonikaClient."""

    tailscale_config_response = device_doesnt_support_skipper(
        create_client().tailscale.get_config()
    )
    assert tailscale_config_response.success
    assert tailscale_config_response.data


def test_client_tailscale_get_status():
    """Test the Tailscale status functionality of the PonikaClient."""

    tailscale_status_response = device_doesnt_support_skipper(
        create_client().tailscale.get_status()
    )
    assert tailscale_status_response.success
    assert tailscale_status_response.data


def test_client_wireless_interfaces_get_status():
    """Test the wireless interfaces status functionality of the PonikaClient."""

    wireless_interfaces_response = device_doesnt_support_skipper(
        create_client().wireless.interfaces.get_status()
    )
    assert wireless_interfaces_response.success
    assert wireless_interfaces_response.data


def test_client_internet_connection_get_status():
    """Test the internet connection status functionality of the PonikaClient."""

    internet_status_response = device_doesnt_support_skipper(
        create_client().internet_connection.get_status()
    )
    assert internet_status_response.success
    assert internet_status_response.data



from typing import Literal, Optional
from pydantic import BaseModel
from typing import TYPE_CHECKING

from ponika.models import ApiResponse

if TYPE_CHECKING:
    from ponika import PonikaClient


class GpsEndpoint:
    def __init__(self, client: "PonikaClient") -> None:
        self.client: "PonikaClient" = client
        self.position = self.GpsPositionEndpoint(client)

    class GetGlobalResponseData(BaseModel):
        """Data model for GET /gps/global response."""

        enabled: Literal["0", "1"]
        galileo_sup: Literal["0", "1"]
        glonass_sup: Literal["0", "1"]
        beidou_sup: Literal["0", "1"]
        dpo_enabled: Optional[Literal["0", "1"]] = None
        mode: Optional[Literal["0", "1"]] = None
        interval: Optional[str] = None
        timeout: Optional[str] = None

    def get_global(self) -> "ApiResponse[GetGlobalResponseData]":
        """Fetch global GPS config."""
        return ApiResponse[self.GetGlobalResponseData].model_validate(
            self.client._get("/gps/global")
        )

    class GpsStatusResponseData(BaseModel):
        """Data model for GET /gps/status response."""

        dpo_support: bool
        uptime: int

    def get_status(self) -> "ApiResponse[GpsStatusResponseData]":
        """Fetch GPS status from the device."""
        return ApiResponse[self.GpsStatusResponseData].model_validate(
            self.client._get("/gps/status")
        )

    class GpsPositionEndpoint:
        def __init__(self, client: "PonikaClient") -> None:
            self.client: "PonikaClient" = client

        class GpsPositionResponseData(BaseModel):
            """Data model for GET /gps/position/status response."""

            accuracy: str
            fix_status: str
            altitude: str
            speed: str
            timestamp: str
            satellites: str
            longitude: str
            latitude: str
            angle: str
            utc_timestamp: str

        def get_status(self) -> "ApiResponse[GpsPositionResponseData]":
            """Fetch GPS position status from the device."""

            return ApiResponse[self.GpsPositionResponseData].model_validate(
                self.client._get("/gps/position/status")
            )



from pydantic import BaseModel, Field
from typing import TYPE_CHECKING

from ponika.models import ApiResponse

if TYPE_CHECKING:
    from ponika import PonikaClient


class InternetConnectionEndpoint:
    def __init__(self, client: "PonikaClient") -> None:
        self.client: "PonikaClient" = client

    class InternetStatusResponseData(BaseModel):
        """Data model for Internet status response."""

        ipv4_status: str = Field(
            description="Indicates whether the device is connected to the internet."
        )
        ipv6_status: str = Field(
            description="IPv6 address of the device, if available."
        )
        dns_status: str = Field(description="DNS status of the device, if available.")

    def get_status(self) -> "ApiResponse[InternetStatusResponseData]":
        """Fetch Internet status from the device."""
        return ApiResponse[self.InternetStatusResponseData].model_validate(
            self.client._get("/internet_connection/status")
        )



from pydantic import BaseModel
from typing import TYPE_CHECKING, List, Optional

from ponika.models import ApiResponse

if TYPE_CHECKING:
    from ponika import PonikaClient


class WirelessEndpoint:
    def __init__(self, client: "PonikaClient") -> None:
        self.client: "PonikaClient" = client
        self.interfaces = self.InterfacesEndpoint(client)

    class InterfacesEndpoint:
        def __init__(self, client: "PonikaClient") -> None:
            self.client: "PonikaClient" = client

        class InterfacesResponseDataItem(BaseModel):
            """Data model for wireless interfaces response."""

            class Client(BaseModel):
                """Data model for wireless client information."""
                expires: Optional[int] = None
                band: str
                ipaddr: Optional[str] = None
                hostname: Optional[str] = None
                tx_rate: int
                macaddr: str
                rx_rate: int
                signal: str
                interface: Optional[str] = None
                device: str

            wifi_id: str
            ifname: str
            disabled: bool
            encryption: str
            vht_supported: Optional[bool] = None # differs from docs says is required
            bssid: str
            num_assoc: int
            clients: List[Client] = []
            status: str
            mode: Optional[str] = None
            multiple: Optional[bool] = None
            # ht_supported: bool
            ssid: str
            id: str
            # conf_id: str
            # auth_status: int



        def get_status(self) -> "ApiResponse[list[InterfacesResponseDataItem]]":
            """Fetch wireless interfaces status from the device."""
            return ApiResponse[list[self.InterfacesResponseDataItem]].model_validate(
                self.client._get("/wireless/interfaces/status")
            )
